#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC
from typing import Coroutine
from pkg_resources import iter_entry_points


from nauti.igather import igather
from nauti.entrypoints import NAUTI_EP_SOURCES
from nauti.config import get_config


__all__ = ["Source", "get_source"]


class Source(ABC):
    name = None
    client_class = None

    async def login(self, *vargs, **kwargs):
        raise NotImplementedError()

    async def logout(self):
        raise NotImplementedError()

    @property
    def is_connected(self):
        raise NotImplementedError()

    @staticmethod
    async def update(updates, callback, creator):

        tasks = dict()
        callback = callback or (lambda _k, _t: True)

        for key, value in updates.items():
            if (coro := creator(key, value)) is None:
                continue

            if not isinstance(coro, Coroutine):
                raise RuntimeError("Source.update requires a coroutine")

            tasks[coro] = (key, value)

        async for orig_coro, res in igather(tasks, limit=100):
            item = tasks[orig_coro]
            callback(item, res)


def get_source(name: str, source_name=None, **kwargs) -> Source:
    """
    Return the Source class designated by the 'name' field.  If a source by `name` is not
    found, then raise RuntimeError.

    Parameters
    ----------
    name: str
        The name of the source type, for example "netbox" or "ipfabric".

    source_name: str
        The specific instance name for the source, as would be found
        in the Config.sources structure.  If `instance_name` is not
        provided, then the `default` config is used.

    """
    if (ep := next(iter_entry_points(NAUTI_EP_SOURCES, name), None)) is None:
        raise RuntimeError(f"ERROR:NOT-FOUND: nauti source: {name}")

    cfg = get_config()
    src_cfg = cfg.sources[name]
    src_inst_cfg = src_cfg.copy()
    cls = ep.load()
    return cls(source_config=src_inst_cfg, **kwargs)



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import unicodedata

__all__ = ['slugify']


def slugify(value, allow_unicode=False):
    """
    NOTE: lifed from django.utils.text.
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower()).strip()
    return re.sub(r"[-\s]+", "-", value)



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from pathlib import Path
from importlib.machinery import SourceFileLoader


NAUTI_EP_SOURCES = 'nauti.sources'
NAUTI_EP_COLLECTIONS = 'nauti.collections'


def find_collection_entrypoints(path: str) -> List[str]:
    from nauti.collection import CollectionMixin
    eps = list()

    for mod_fp in Path(path).glob('[!_]*.py'):
        mod_pxfp = str(mod_fp.as_posix())
        mod_name = mod_pxfp.replace('/', '.')
        mod = SourceFileLoader(mod_name, mod_pxfp).load_module()
        for exported in getattr(mod, '__all__', []):
            cls = getattr(mod, exported)
            if issubclass(cls, CollectionMixin) and cls != CollectionMixin:
                eps.append(f"{cls.name} = {mod.__package__}:{exported}")

    return eps


def find_source_entrypoints(path: str) -> List[str]:
    from nauti.source import Source
    eps = list()

    for mod_fp in Path(path).glob('[!_]*.py'):
        mod_pxfp = str(mod_fp.as_posix())
        mod_name = mod_pxfp.replace('/', '.')
        mod = SourceFileLoader(mod_name, mod_pxfp).load_module()

        for exported in getattr(mod, '__all__', []):
            cls = getattr(mod, exported)
            if issubclass(cls, Source) and cls != Source:
                eps.append(f"{cls.name} = {mod.__package__}:{exported}")

    return eps



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This file contains the filtering functions that are using to process the
'--include' and '--exclude' command line options.  The code in this module is
not specific to the netcfgbu source_records column names, can could be re-used
for other CSV related tools and use-cases.
"""

from typing import List, AnyStr, Optional, Callable, Dict, Sequence
import re
import operator


__all__ = ["create_filter"]


value_pattern = r"(?P<value>\S+)$"
wordsep_re = re.compile(r"\s+|,")


def mk_op_filter(_reg, _fieldn):
    """ create a single op filter """

    def op_filter(rec):
        """ using the regular expression match call """
        return _reg.match(rec[_fieldn])

    op_filter.__doc__ = f"limit_{_fieldn}({_reg.pattern})"
    op_filter.__name__ = op_filter.__doc__
    op_filter.__qualname__ = op_filter.__doc__

    return op_filter


def create_filter_function(op_filters, optest_fn):
    """ create a filtering functions based on the operational test """

    def filter_fn(rec):
        """ for each of the filters """
        for op_fn in op_filters:
            if optest_fn(op_fn(rec)):
                return False

        return True

    return filter_fn


def create_filter(
    constraints: List[AnyStr],
    field_names: Sequence[AnyStr],
    include: Optional[bool] = True,
) -> Callable[[Dict], bool]:
    """
    This function returns a function that is used to filter source_records records.

    Parameters
    ----------
    constraints:
        A list of contraint expressions that are in the form "<field-name>=<value>".

    field_names:
        A list of known field names

    include:
        When True, the filter function will match when the constraint is true,
        for example if the contraint is "os_name=eos", then it would match
        records that have os_name field euqal to "eos".

        When False, the filter function will match when the constraint is not
        true. For exampl if the constraint is "os_name=eos", then the filter
        function would match recoreds that have os_name fields not equal to
        "eos".

    Returns
    -------
    The returning filter function expects an source_records record as the single
    input parameter, and the function returns True/False on match.
    """
    fieldn_pattern = "^(?P<keyword>" + "|".join(fieldn for fieldn in field_names) + ")"
    field_value_reg = re.compile(fieldn_pattern + "=" + value_pattern)

    op_filters = list()
    for filter_expr in constraints:

        # next check for keyword=value filtering use-case

        if (mo := field_value_reg.match(filter_expr)) is None:
            raise ValueError(f"Invalid filter expression: {filter_expr}")

        fieldn, value = mo.groupdict().values()

        try:
            value_reg = re.compile(f"^{value}$", re.IGNORECASE)

        except re.error as exc:
            raise ValueError(
                f"Invalid filter regular-expression: {filter_expr}: {str(exc)}"
            )

        op_filters.append(mk_op_filter(value_reg, fieldn))

    optest_fn = operator.not_ if include else operator.truth
    filter_fn = create_filter_function(op_filters, optest_fn)
    filter_fn.op_filters = op_filters
    filter_fn.constraints = constraints

    return filter_fn



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

import os
from typing import TextIO
from contextvars import ContextVar
from pathlib import Path

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

import toml
from pydantic import ValidationError
from pydantic_env import config_validation_errors


# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from .config_models import ConfigModel
from nauti import consts

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------


__all__ = ["get_config", "load_config_file", "load_default_config_file", "ConfigModel"]

# -----------------------------------------------------------------------------
#
#                              CODE BEGINS
#
# -----------------------------------------------------------------------------

g_config = ContextVar("config")


def get_config() -> ConfigModel:
    return g_config.get()


def load_config_file(filepath: TextIO):
    # as_fp = Path(filepath.name)
    # fp_dir = as_fp.parent
    # cfg_obj = dict()

    try:
        cfg_obj = toml.load(filepath)
        cfg_obj['config_file'] = filepath.name
    except ValueError as exc:
        raise RuntimeError(
            f"FAIL: loading file {str(exc)}"
        )

    # try:
    #     for each_fp in fp_dir.glob('*.toml'):
    #         cfg_obj.update(toml.load(each_fp.open()))
    # except Exception as exc:
    #     raise RuntimeError(f"FAIL: loading file {str(exc)}")

    try:
        config_obj = ConfigModel.parse_obj(cfg_obj)
        g_config.set(config_obj)
        return config_obj

    except ValidationError as exc:
        raise RuntimeError(
            config_validation_errors(errors=exc.errors(), filepath=filepath.name)
        )


def load_default_config_file():
    cfg_file = os.environ.get(consts.ENV_CONFIG_FILE, consts.DEFAULT_CONFIG_FILE)
    load_config_file(filepath=open(cfg_file))



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from functools import partial, lru_cache
import re
from operator import concat

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from nauti.config import get_config

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = ["normalize_hostname"]


# -----------------------------------------------------------------------------
#
#                              CODE BEGINS
#
# -----------------------------------------------------------------------------


@lru_cache()
def domain_remover():
    cfg_obj = get_config()
    any_domain = "|".join(
        map(re.escape, map(partial(concat, "."), cfg_obj.defaults.domain_names))
    )
    return partial(re.compile(any_domain).sub, repl="")


def normalize_hostname(hostname):
    return domain_remover()(string=hostname.lower())



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# The `igather` async generator was found here:
#   https://bugs.python.org/issue30782
#   Author: Andrey Paramonov (aparamon)
#
# in a discussion on how to treat a large collection of tasks.  I've modified the code
# slightly so that I get back the original coroutine; but beyond that it is as it was found.


import asyncio
import math

__all__ = ["igather"]


async def igather(coros, limit=None):
    coros = iter(coros)

    buf = asyncio.Queue()
    sem = asyncio.Semaphore(limit or math.inf)

    async def submit(_coros, _buf):
        while True:
            await sem.acquire()
            try:
                # TODO: additionally support async iterators
                _coro = next(_coros)
            except StopIteration:
                break
            _task = asyncio.create_task(_coro)
            _buf.put_nowait(_task)
        await _buf.put(None)

    async def consume(_buf):
        while True:
            _task = await _buf.get()
            if _task:
                v = await asyncio.wait_for(_task, None)
                sem.release()
                yield _task.get_coro(), v  # the yield will be Tuple(original-coro, task-result)
            else:
                break

    submit_task = asyncio.create_task(submit(coros, buf))

    try:
        async for result in consume(buf):
            yield result

    except Exception as exc:  # noqa
        submit_task.cancel()
        # cancel scheduled
        while not buf.empty():
            task = buf.get_nowait()
            if task:
                task.cancel()
                try:
                    await task
                except Exception:
                    pass

        # cancel pending
        for coro in coros:
            asyncio.create_task(coro).cancel()

        raise


async def iawait(coros, limit=None):
    async for _ in igather(coros, limit):
        pass



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Callable, Optional, Iterable
from collections import namedtuple

from nauti.collection import Collection

DiffResults = namedtuple("DiffResults", ["missing", "extras", "changes"])


def diff(
    source_from: Collection,
    sync_to: Collection,
    fields: Optional[Iterable] = None,
    fields_cmp: Optional[Dict[str, Callable]] = None,
):
    """
    The source and other collections have already been fetched, fingerprinted, and keyed.
    This method is used to create a diff report so that action can be taken to account for
    the differences.

    Parameters
    ----------
    source_from:
        The collection that is the source of truth for the diff

    sync_to:
        The collection that represents the destination of the update

    fields:
        A list of field names used for comparison purposes.  If not provided,
        then all fields in the collection FIELDS attribute will be used.  A
        Caller can provide `fields` when they want to diff only a subset of the
        collection FIELDS.

    fields_cmp:
        Dictionary mapping the field name to a function used to "normalize" the
        value so that it can be compared.  A common function would be
        `str.lower` to convert a field (hostname) to lower for comparison
        purposes.

    Returns
    -------
    DiffResults:
        missing: Dict[Tuple]
        changes: List[Tuple[Dict, Dict]]
    """
    sync_to_keys = set(sync_to.inventory)
    source_from_keys = set(source_from.inventory)

    missing_keys = source_from_keys - sync_to_keys
    extra_keys = sync_to_keys - source_from_keys
    shared_keys = source_from_keys & sync_to_keys

    # missing key dict; key=source_records-key, value=key-fingerprint
    missing_key_items = {key: source_from.inventory[key] for key in missing_keys}
    extra_key_items = {key: sync_to.inventory[key] for key in extra_keys}

    changes = dict()

    if not fields_cmp:
        fields_cmp = {}

    for field in (fields or source_from.FIELDS):
        if field not in fields_cmp:
            fields_cmp[field] = lambda f: f

    for key in shared_keys:
        source_fp = source_from.inventory[key]
        sync_fp = sync_to.inventory[key]

        item_changes = dict()

        for field, field_fn in fields_cmp.items():
            if field_fn(source_fp[field]) != field_fn(sync_fp[field]):
                item_changes[field] = source_fp[field]

        if len(item_changes):
            changes[key] = item_changes

    if not any((missing_key_items, extra_key_items, changes)):
        return None

    return DiffResults(
        missing=missing_key_items, changes=changes, extras=extra_key_items
    )



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional, List, Dict, Any, Union
from pathlib import Path

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

from pydantic import (
    BaseModel, Extra,
    validator, Field
)
from pydantic_env.models import (
    NoExtraBaseModel, EnvSecretStr,
    EnvUrl
)

import toml

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = [
    'ConfigModel',
    'TokenCredential', 'ClientCredential',
    'SourcesModel'
]

# -----------------------------------------------------------------------------
#
#                              CODE BEGINS
#
# -----------------------------------------------------------------------------


class TokenCredential(NoExtraBaseModel):
    token: EnvSecretStr


class ClientCredential(NoExtraBaseModel):
    client_id: EnvSecretStr
    client_secret: EnvSecretStr


AnyCredentialsModel = Union[TokenCredential, ClientCredential]


# -----------------------------------------------------------------------------
#
#                                Sources
#
# -----------------------------------------------------------------------------

class SourceInstanceModel(BaseModel):
    """ A specific source instance, as there could be more than one """
    url: EnvUrl
    credentials: AnyCredentialsModel
    options: Optional[Dict[str, Any]] = Field(default_factory=dict)


class SourcesModel(BaseModel):
    default: SourceInstanceModel
    vars: Optional[Dict[str, EnvSecretStr]]

# -----------------------------------------------------------------------------
#
#                                Collections
#
# -----------------------------------------------------------------------------


class CollectionsModel(BaseModel):

    # The BaseModel defines the attribute `fields` so we are not suppoed to use
    # that by default.  The "workaround" is to alias a different attribute name,
    # in this case using `fields_`, and then override the `fields` property. The
    # pydantic library depcreciated the use of `fields` in favor of `__fields__`
    # so I presume this workaround is OK.

    fields_: Dict[str, Any] = Field(alias='fields')
    maps: Dict[str, Dict]

    @property
    def fields(self):
        return self.fields_


class ConfigModel(BaseModel):
    class Config:
        extra = Extra.allow

    config_file: str
    domain_names: Optional[List[str]]
    sources: Dict[str, SourcesModel]
    collections: Dict[str, CollectionsModel]

    @validator('sources', pre=True)
    def _each_source(cls, v, values):
        par_dir = Path(values['config_file']).parent
        return {name: _load_item_config(par_dir, name, SourcesModel) for name in v}

    @validator('collections', pre=True)
    def _each_collection(cls, v, values):
        par_dir = Path(values['config_file']).parent
        return {name: _load_item_config(par_dir, name, CollectionsModel) for name in v}


# -----------------------------------------------------------------------------
#
#                              PRIVATE FUNCTIONS
#
# -----------------------------------------------------------------------------

def _load_item_config(cfg_dir, item_name, item_cls):
    file_p = cfg_dir.joinpath(item_name + ".toml")
    content = toml.load(file_p.open())
    name = next(iter(content))
    return item_cls.parse_obj(content[name])



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import List, Dict, Any, Callable, Tuple, Optional, Type
from abc import ABC
from operator import itemgetter
from pkg_resources import iter_entry_points

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

from bidict import bidict

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from nauti.log import get_logger
from nauti.source import Source
from nauti.entrypoints import NAUTI_EP_COLLECTIONS
from nauti.config import get_config

__all__ = ["Collection", "CollectionMixin", "CollectionCallback", "get_collection"]

CollectionCallback = Type[Callable[[Tuple, Any], None]]


class CollectionMixin(object):

    # The name of the collection, used when calling `get_collection`
    name = None

    # The source class bound to the collection, providing source-specific
    # collection implementations.
    source_class = None

    # List of expected fields collection provides
    FIELDS = None

    # Default set of field names used to create keys.
    KEY_FIELDS = None

    # List of additional fields that the collection makes avaiable.
    ADDNL_FIELDS = None

    # List of fields that collection does not provide, i.e. used to declare that
    # the collection provides a subset of FIELDS.
    NO_FIELDS = None


class Collection(ABC, CollectionMixin):

    # -------------------------------------------------------------------------
    #
    #                     Required Subclass Methods
    #
    # -------------------------------------------------------------------------

    async def fetch(self, **fetch_args):
        """ retrieve source records from the collection provider """
        raise NotImplementedError()

    # async def fetch_keys(self, keys: Dict):
    #     raise NotImplementedError()

    def itemize(self, rec: Dict) -> Dict:
        """ Translate a source specific record into a dict of field items """
        raise NotImplementedError()

    async def add_items(
        self, items: Dict, callback: Optional[CollectionCallback] = None
    ):
        """ add missing `items` to the collection """
        raise NotImplementedError()

    async def update_items(
        self, items: Dict, callback: Optional[CollectionCallback] = None
    ):
        """ update the existing `items` with field changes """
        raise NotImplementedError()

    async def delete_items(
        self, items: Dict, callback: Optional[CollectionCallback] = None
    ):
        """ remove the `items` from the collection """
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    #
    #                     Class Init
    #
    # -------------------------------------------------------------------------

    def __init__(self, source: Source):

        # `source_records` is a list of recoreds as they are obtained from the
        # source.  The structure each source_records record is specific to the
        # source.  The inentory record is "fingerprinted" for collection fields;
        # which in turn are used to create inventory.

        self.source_records: List[Any] = list()

        # `inventory` is a dict where the key=<fields-key> and the value is the
        # fields-record of the source record.

        self.inventory: Dict[Tuple, Dict] = dict()

        # `source_record_keys` is a dict key=<fields-key>, value=<source-record>
        # that is used to cross reference the fields-key to a source specific
        # record ID which is typically found in the source specific response
        # record. The uid value is used when making updates to an exists record
        # in the source.

        self.source_record_keys: Dict[Tuple, Any] = dict()

        # The Source instance providing connectivity for the Collection
        # processing.

        self.source = source

        # `cache` is expected to be used by the subclass to store information
        # that it may need across various calls; for example caching device
        # records that may have information required by processing other
        # collections (ipaddrs).

        self.cache = dict()

        # `config` is the Config.collections[<name>] structure, initialied in
        # the call to get_collection().
        self.config = None

        # `mappings` ihe Config.collections[<name>].maps[<source-name>]
        # structure, initialized in the call to get_collection().
        self.maps = dict()

    def make_keys(
        self,
        *fields,
        with_filter: Optional[Callable[[Dict], bool]] = None,
        with_translate=None,
        with_inventory=None,
    ):
        if not len(self.source_records):
            get_logger().info(
                f"Collection {self.name}:{self.source_class.__name__}: inventory empty."
            )
            return

        with_filter = with_filter if with_filter else lambda x: True
        with_translate = with_translate or (lambda x: x)

        kf_getter = itemgetter(*(fields or self.KEY_FIELDS))

        if not with_inventory:
            self.inventory.clear()

        for rec in with_inventory or self.source_records:
            try:
                item = self.itemize(rec)

                if not with_filter(item):
                    continue

            except Exception as exc:
                raise RuntimeError("itimized failed", rec, exc)

            as_key = with_translate(kf_getter(item))
            self.inventory[as_key] = item
            self.source_record_keys[as_key] = rec

    # @classmethod
    # def get_collection(cls, source, name) -> "Collection":
    #     try:
    #         c_cls = next(
    #             iter(
    #                 c_cls
    #                 for c_cls in cls.__subclasses__()
    #                 if all(
    #                     (
    #                         c_cls.name == name,
    #                         c_cls.source_class,
    #                         isinstance(source, c_cls.source_class),
    #                     )
    #                 )
    #             )
    #         )
    #
    #     except StopIteration:
    #         raise RuntimeError(
    #             f"NOT-FOUND: Collection {name} for source class: {source.name}"
    #         )
    #
    #     return c_cls(source=source)

    def __len__(self):
        return len(self.source_records)


def get_collection(source, name: str) -> Collection:

    for ep in iter_entry_points(NAUTI_EP_COLLECTIONS, name):
        cls = ep.load()
        if isinstance(source, cls.source_class):
            break
    else:
        raise RuntimeError(f"ERROR:NOT-FOUND: nauti collection: {name}")

    cfg = get_config()
    col_obj: Collection = cls(source=source)
    col_obj.config = cfg.collections[name]

    if (col_maps := col_obj.config.maps.get(source.name)) is not None:
        for map_name, map_data in col_maps.items():
            col_obj.maps[map_name] = bidict(
                {key: value or key.lower() for key, value in map_data.items()}
            )

    return col_obj



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from setuptools import setup, find_packages

package_name = "nauti"
package_version = open("VERSION").read().strip()


def requirements(filename="requirements.txt"):
    return open(filename.strip()).readlines()


with open("README.md", "r") as fh:
    long_description = fh.read()


# -----------------------------------------------------------------------------
#
#                                 Main Setup
#
# -----------------------------------------------------------------------------

setup(
    name=package_name,
    version=package_version,
    description="Network Automation Tools Integrator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jeremy Schulman",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements(),
    entry_points={
        "console_scripts": ["nauti = nauti.cli.__main__:script"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Networking",
    ],
)



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Callable
from functools import lru_cache, partial
import re

from nauti.config import get_config


@lru_cache()
def _expaner_os() -> Callable[[str], str]:
    config = get_config()
    cfg_map = config.maps["interfaces"]
    mapper = re.compile(r"|".join(list(cfg_map)))

    def _expander(ifname):
        return mapper.sub(lambda mo: cfg_map[mo.group(0)], ifname)

    return _expander


def expand_interface(ifname: str) -> str:
    return _expaner_os()(ifname)


@lru_cache()
def domain_remover():
    cfg_obj = get_config()
    any_domain = "|".join(
        re.escape(f".{domain}") for domain in cfg_obj.defaults.domain_names
    )
    return partial(re.compile(any_domain).sub, repl="")


def normalize_hostname(hostname):
    return domain_remover()(string=hostname.lower())



#      Copyright (C) 2020  Jeremy Schulman
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from nauti.collection import CollectionMixin

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = ["DeviceCollection"]


class DeviceCollection(CollectionMixin):
    name = "devices"

    FIELDS = (
        "sn",
        "hostname",
        "ipaddr",
        "site",
        "os_name",
        "vendor",
        "model",
    )

    KEY_FIELDS = ("sn",)
